
{*******************************************************}
{                                                       }
{       Borland TeamSource                              }
{                                                       }
{       Copyright (c) 1999 Inprise Corporation          }
{                                                       }
{*******************************************************}

unit ZLibIVAPI;

interface

{
  ZLib archive files are a simple sequential storage system.
  Each new version is stored in it's entirety, and no deltas
  are generated.

  Layout of a ZLib archive file:

  ZHeader
    Signature        - [11] "ZLIBARCHIVE"
    Version          - [4] file format version
    DirectoryOffset  - [4] offset from start of file to ZDirectory
    RevCount         - [4] number of revisions in ZData
    OriginalExt      - [4] file's original extension
    Creator          - [30] name of creator
    CreationDate     - [TDateTime] when the file was created

  ZData
    Raw file data

  ZDirectory
    The directory comes in two formats. The 1.0 format and the 2.0 format.
    The format is determined by the Version field in the header.

  Directory format:

     ZDirectory
       Entries        - [RevCount*SizeOf(ZDirEntry)] RevCount entries of ZDirEntry

     ZDirEntry
        Version       - [ZVersion] Version stored as a ZVersion
        Offset        - [4] within the ZData section
        OriginalSize  - [4] file's original size
        StoredSize    - [4] file's size in ZData section
        FileDate      - [4] file's date
        CheckinDate   - [4] date added to archive
        Author        - [string] author
        Comment       - [string] change comments

     ZVersion
       BranchCount    - [2] up to 65535 nested branches
       Version        - [BranchCount*4] Each word pair represents the major/minor
                        revisions
}

uses Classes, Windows, IVAPI, StdIVAPI, SysUtils;

type
  TZLibArchive = class;
  
  ZUserName = string[30];
  ZFileExt = string[64];
  ZFileSignature = array[0..10] of Char;

  PBranch = ^ZBranch;
  ZBranch = record
    Major: Word;
    Minor: Word;
  end;

  ZRevision = type string; { Stores Length(ZRevision) / 4 ZBranch's }

  ZHeader = packed record
    Signature: ZFileSignature;
    Version: Integer;
    DirectoryOffset: Integer;
    RevCount: Integer;
    OriginalExt: ZFileExt;
    Creator: ZUserName;
    CreationDate: TDateTime;
  end;

  TZDirEntry = class
    FormatVersion: Integer;
    Revision: ZRevision;
    Offset: Longint;
    OriginalSize: Longint;
    StoredSize: Longint;
    FileDate: TDateTime;
    CheckinDate: TDateTime;
    Author: string;
    Comment: string;
    Locked: Boolean;
    constructor Create(AFormatVersion: Integer);
    procedure LoadFromStream(S: TStream);
    procedure SaveToStream(S: TStream);
  end;

  { TZLibVersionController }
  TZLibVersionController = class(TInterfacedObject, IVersionController)
  private
    FUser: IUser;

    { IVersionController }
    function GetUser: string;
    function OpenArchive(const ArchiveName: string): IArchive;
    function CreateArchive(const ArchiveName: string; const SourceName: string;
      const Description: string): IArchive;
    function GetArchiveName(const SourceName: string): string;
    function GetArchiveAge(const ArchiveName: string): Integer;
    function ArchiveExists(const ArchiveName: string): Boolean;
    function IsArchive(const SourceName: string): Boolean;
    function GetRevisionHistory(const ArchiveName: string): IRevisionHistory;
    function GetFileMerger: IFileMerger;
    function GetOptions: IControllerOptions;
  protected
    function InternalOpenArchive(const ArchiveName: string): TZLibArchive;
  public
    constructor Create(User: IUser); 
  end;

  TZLibArchive = class(TInterfacedObject, IArchive)
  private
    FStream: TStream;
    FOpen: Boolean;
    FReadOnly: Boolean;
    FController: TZLibVersionController;
    FArchiveName: string;
    FSourceName: string;
    FHeader: ZHeader;
    FDirectory: TList;
    FCurrentRevision: TZDirEntry;
    FNewRevision: ZRevision;

    { IArchive }
    function GetAuthor: string;
    function GetComment: string;
    function GetDescription: string;
    function GetLocked: Boolean;
    function GetLockedBy: string;
    function GetSourceDate: TDateTime;
    function GetSourceName: string;
    function GetVersion: string;
    procedure PositionToVersion(const Version: string);
    procedure SetLocked(Value: Boolean);
    function Checkin(const Source: ISource; const Comment: string): string;
    function CheckOut: ISource;
    function Source: ISource;
    procedure CopyToFile(const Target: string);


    procedure OpenStream(Mode: Integer);
    procedure OpenArchive;
    procedure CreateArchive;
    procedure ValidState;

    procedure LoadDirectoryInfo;
    procedure ClearDirectory;

    procedure WriteHeader;
    procedure WriteDirectory;
    function IndexOfRevision(Revision: ZRevision): Integer;

    function GetCurrentRevision: ISource;

    function GetCreator: string;
    function GetCreationDate: TDateTime;
    function GetDirectoryCount: Integer;
    function GetDirectoryEntry(I: Integer): TZDirEntry;
    function GetNextRevision(ForceBranch: Boolean): ZRevision;

    function GetZLibBranchTip(Revision: ZRevision): ZRevision;
    function PutSource(Source: ISource; const Comment: string;
      ForceBranch: Boolean): string;
  public
    constructor Create(Controller: TZLibVersionController;
      const ArchiveName, SourceName: string; Existing: Boolean);
    destructor Destroy; override;

    // report properties
    property ArchiveName: string read FArchiveName;
    property SourceName: string read GetSourceName;
    property Creator: string read GetCreator;
    property CreationDate: TDateTime read GetCreationDate;
    property DirectoryCount: Integer read GetDirectoryCount;
    property DirectoryEntry[I: Integer]: TZDirEntry read GetDirectoryEntry;
  end;

  TZRevRecord = class
    Revision: string;
    ModDate: TDateTime;
    CheckinDate: TDateTime;
    Author: string;
    Comment: string;
    OriginalSize: Integer;
    StoredSize: Integer;
  end;

  TZLibRevisionHistory = class(TInterfacedObject, IRevisionHistory)
  private
    FList: TList;
    FArchiveName: string;
    FSourceName: string;
    FCreationDate: TDateTime;
    FCreator: string;

    { IRevisionHistory }
    function GetArchiveName: string;
    function GetSourceName: string;
    function GetCreationDate: TDateTime;
    function GetCreator: string;
    function GetRevCount: Integer;
    function GetRevision(Index: Integer): string;
    function GetFileSize(Index: Integer): Integer;
    function GetStoredSize(Index: Integer): Integer;
    function GetLastModDate(Index: Integer): TDateTime;
    function GetCheckinDate(Index: Integer): TDateTime;
    function GetAuthor(Index: Integer): string;
    function GetComment(Index: Integer): string;
  public
    constructor Create(Controller: TZLibVersionController;
      const ArchiveName: string);
    destructor Destroy; override;
  end;


  EZArchive = class(Exception);

function TextToZRevision(const S: string): ZRevision;
function ZRevisionToText(Revision: ZRevision): string;

implementation

uses ZLib;

const
  FileSignature = 'ZLIBARCHIVE';
  FileVersion = Integer($FACE0300);

resourcestring
  sInvalidFileSignature = '%s is not a valid ZArchive file';
  sInvalidFileVersion = '%s requires version %s of the ZArchive manager';
  sInvalidVersionFormat = '%s is not in valid ZArchive version format';
  sNotInitialized = 'Archive not initialized';
  sReadOnlyArchive = 'Archive is read-only';

type
  { TSourceAdaptor - a TStream derivitive which adapts an ISource to look
    like a TStream to a TCompressionStream }
  TSourceAdaptor = class(TStream)
  private
    FSource: ISource;
  public
    constructor Create(Source: ISource);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

function GetTemporaryFileName(const Prefix: string): string;
var
  Buffer: array[0..MAX_PATH] of Char;
  Path: string;
begin
  SetString(Path, Buffer, GetTempPath(SizeOf(Buffer), Buffer));
  GetTempFileName(PChar(Path), PChar(Prefix), 0, Buffer);
  Result := Buffer;
end;


{ TSourceAdaptor }
constructor TSourceAdaptor.Create(Source: ISource);
begin
  inherited Create;
  FSource := Source;
end;

function TSourceAdaptor.Read(var Buffer; Count: Longint): Longint;
begin
  Result := FSource.Read(@Buffer, Count);
end;

function TSourceAdaptor.Write(const Buffer; Count: Longint): Longint;
begin
  Result := 0;  // not supported
end;

function TSourceAdaptor.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning: Result := 0;
    soFromCurrent: Result := 0;
    soFromEnd: Result := FSource.GetLength;
    else Result := 0;
  end;
end;


{ general functions }

function TextToZRevision(const S: string): ZRevision;
var
  P: PChar;
  Major, Minor: Integer;
  L: Integer;
begin
  P := Pointer(S);
  Result := '';
  while P^ <> #0 do
  begin
    Major := 0;
    Minor := 0;
    while P^ in ['0'..'9'] do
    begin
      Major := Major * 10 + (Byte(P^) - Byte('0'));
      Inc(P);
    end;
    if (Major = 0) or (P^ <> '.') then
      raise Exception.CreateFmt(sInvalidVersionFormat, [S]);
    Inc(P);
    while P^ in ['0'..'9'] do
    begin
      Minor := Minor * 10 + (Byte(P^) - Byte('0'));
      Inc(P);
    end;
    L := Length(Result);
    SetLength(Result, Length(Result) + SizeOf(ZBranch));
    PBranch(@Result[L + 1])^.Major := Major;
    PBranch(@Result[L + 1])^.Minor := Minor;
    if P^ = '.' then Inc(P);
  end;
end;

function ZRevisionToText(Revision: ZRevision): string;
var
  BranchCount: Integer;
  I: Integer;
begin
  Assert(Length(Revision) mod SizeOf(ZBranch) = 0);
  BranchCount := Length(Revision) div SizeOf(ZBranch);
  Result := '';
  for I := 0 to BranchCount - 1 do
  begin
    with PBranch(@Revision[I * SizeOf(ZBranch) + 1])^ do
      if I = 0 then
        Result := Format('%d.%d', [Major, Minor]) else
        Result := Format('%s.%d.%d', [Result, Major, Minor]);
  end;
end;

procedure Error(const S: string);
begin
  raise EZArchive.Create(S);
end;

function VersionString(Version: Integer): string;
var
  Major: Byte;
  Minor: Byte;
begin
  Major := WordRec(LongRec(Version).Lo).Hi;
  Minor := WordRec(LongRec(Version).Lo).Lo;
  Result := Format('%d.%2d', [Major, Minor]);
end;

{ TZLibVersionController }
constructor TZLibVersionController.Create(User: IUser);
begin
  inherited Create;
  FUser := User;
end;

function TZLibVersionController.InternalOpenArchive(
  const ArchiveName: string): TZLibArchive;
begin
  Result := TZLibArchive.Create(Self, ArchiveName, '', True);
end;


function TZLibVersionController.OpenArchive(const ArchiveName: string): IArchive;
begin
  Result := InternalOpenArchive(ArchiveName);
end;

function TZLibVersionController.CreateArchive(const ArchiveName, SourceName,
  Description: string): IArchive;
begin
  ForceDirectories(ExtractFilePath(ArchiveName));
  Result := TZLibArchive.Create(Self, ArchiveName, SourceName, False);
end;

function TZLibVersionController.GetArchiveName(const SourceName: string): string;
begin
  Result := SourceName;
  if AnsiPos('.', Result) = Length(Result) then
    SetLength(Result, Length(Result) - 1);
  Result := Result + '.z';
end;

function TZLibVersionController.IsArchive(const SourceName: string): Boolean;
var
  Signature: ZFileSignature;
  Stream: TStream;
begin
  Result := False;
  if FileExists(SourceName) then
  begin
    Stream := TFileStream.Create(SourceName, fmOpenRead);
    try
      Stream.Read(Signature, SizeOf(Signature));
      Result := StrLComp(Signature, FileSignature, SizeOf(ZFileSignature)) = 0;
    finally
      Stream.Free;
    end;
  end;
end;

function TZLibVersionController.GetRevisionHistory(const ArchiveName: string): IRevisionHistory;
begin
  Result := TZLibRevisionHistory.Create(Self, ArchiveName);
end;

function TZLibVersionController.GetFileMerger: IFileMerger;
begin
  Result := nil;
end;

function TZLibVersionController.GetOptions: IControllerOptions;
begin
  Result := nil;
end;

function TZLibVersionController.GetUser: string;
begin
  Result := FUser.GetName;
end;

function TZLibVersionController.ArchiveExists(
  const ArchiveName: string): Boolean;
begin
  Result := FileExists(ArchiveName);
end;

function TZLibVersionController.GetArchiveAge(
  const ArchiveName: string): Integer;
begin
  Result := FileAge(ArchiveName);
end;

{ TZLibArchive }
constructor TZLibArchive.Create(Controller: TZLibVersionController;
  const ArchiveName, SourceName: string; Existing: Boolean);
var
  TrunkTipRev: ZRevision;
  TrunkTipIndex: Integer;
begin
  inherited Create;
  FOpen := False;
  FArchiveName := ArchiveName;
  FSourceName := SourceName;
  FController := Controller;

  FDirectory := TList.Create;

  if Existing then OpenArchive
  else CreateArchive;
  FNewRevision := #1#0#0#0;

  // move to the tip of the trunk by default
  if FCurrentRevision <> nil then
  begin
    TrunkTipRev := GetZLibBranchTip(FCurrentRevision.Revision);
    TrunkTipIndex := IndexOfRevision(TrunkTipRev);
    if TrunkTipIndex <> -1 then
      FCurrentRevision := FDirectory[TrunkTipIndex];
  end;
end;

destructor TZLibArchive.Destroy;
begin
  FStream.Free;
  ClearDirectory;
  FDirectory.Free;
  inherited Destroy;
end;

procedure TZLibArchive.ClearDirectory;
var
  I: Integer;
begin
  for I := 0 to FDirectory.Count - 1 do
    TZDirEntry(FDirectory[I]).Free;
  FDirectory.Clear;
end;

procedure TZLibArchive.OpenStream(Mode: Integer);
begin
  try
    FStream := TFileStream.Create(FArchiveName, Mode);
  except
    on E: EFOpenError do
    begin
      // may have occurred due to File Not Found, or Read-Only access
      if FileExists(FArchiveName) and ((fmOpenReadWrite and Mode) <> 0) then
      begin
        Mode := (Mode and $F0) or fmOpenRead;  // preserve share flags
        FStream := TFileStream.Create(FArchiveName, Mode);
        FReadOnly := True;
      end
      else raise;
    end
    else raise;
  end;
end;

procedure TZLibArchive.OpenArchive;
begin
  OpenStream(fmOpenReadWrite);

  // read the header information
  FStream.Read(FHeader.Signature, SizeOf(FHeader.Signature));
  if StrComp(FHeader.Signature, FileSignature) <> 0 then
    Error(Format(sInvalidFileSignature, [FArchiveName]));

  FStream.Read(FHeader.Version, SizeOf(FHeader.Version));
  if FHeader.Version <> FileVersion then
    Error(Format(sInvalidFileVersion, [FArchiveName,
      VersionString(FHeader.Version)]));

  with FStream, FHeader do
  begin
    Read(DirectoryOffset, SizeOf(DirectoryOffset));
    Read(RevCount, SizeOf(RevCount));
    Read(OriginalExt, SizeOf(OriginalExt));
    Read(Creator, SizeOf(Creator));
    Read(CreationDate, SizeOf(CreationDate));
  end;

  // read directory information
  LoadDirectoryInfo;

  FOpen := True;
end;

procedure TZLibArchive.CreateArchive;
begin
  OpenStream(fmCreate);

  with FHeader do
  begin
    StrCopy(Signature, FileSignature);
    Version := FileVersion;
    DirectoryOffset := SizeOf(FHeader);
    RevCount := 0;
    OriginalExt := ExtractFileExt(FSourceName);
    Creator := FController.GetUser;
    CreationDate := Now;
  end;
  FStream.Write(FHeader, SizeOf(FHeader));

  FOpen := True;
end;

procedure TZLibArchive.ValidState;
begin
  if not FOpen then
    Error(sNotInitialized);
end;

function TZLibArchive.IndexOfRevision(Revision: ZRevision): Integer;
var
  Entry: TZDirEntry;
begin
  for Result := 0 to FDirectory.Count - 1 do
  begin
    Entry := FDirectory[Result];
    if Entry.Revision = Revision then
      Exit;
  end;
  Result := -1;
end;

procedure TZLibArchive.LoadDirectoryInfo;
var
  DirEntry: TZDirEntry;
//  Prior: TZDirEntry;
  I: Integer;
begin
  ClearDirectory;
  FStream.Position := FHeader.DirectoryOffset;
  for I := 0 to FHeader.RevCount - 1 do
  begin
    DirEntry := TZDirEntry.Create(FHeader.Version);
    try
      DirEntry.LoadFromStream(FStream);
      FDirectory.Add(DirEntry);
    except
      DirEntry.Free;
      raise;
    end;
  end;
  if FDirectory.Count > 0 then FCurrentRevision := FDirectory[0];
end;

function TZLibArchive.GetAuthor: string;
begin
  ValidState;
  if FCurrentRevision <> nil then Result := FCurrentRevision.Author
  else Result := '';
end;

function TZLibArchive.GetComment: string;
begin
  ValidState;
  if FCurrentRevision <> nil then Result := FCurrentRevision.Comment
  else Result := '';
end;

function TZLibArchive.GetDescription: string;
begin
  ValidState;
  Result := '';
  { !!!  doesn't support file descriptions }
end;

function TZLibArchive.GetLocked: Boolean;
begin
  ValidState;
  if FCurrentRevision <> nil then Result := FCurrentRevision.Locked
  else Result := False;
end;

function TZLibArchive.GetLockedBy: string;
begin
  if GetLocked then Result := FController.GetUser 
  else Result := '';
end;

function TZLibArchive.GetSourceDate: TDateTime;
begin
  ValidState;
  if FCurrentRevision <> nil then
    Result := LocalDateToUTCDate(FCurrentRevision.FileDate)
  else Result := 0;
end;

function TZLibArchive.GetSourceName: string;
begin
  ValidState;
  Result := ChangeFileExt(ExtractFileName(FArchiveName), '');  // strip off the terminating '.z'
  if ExtractFileExt(Result) = '' then
    Result := ChangeFileExt(Result, FHeader.OriginalExt);
end;

function IsPrefix(const Prefix, Value: string): Boolean;
var
  I: Integer;
begin
  Result := False;
  if Length(Prefix) > Length(Value) then Exit;
  for I := 1 to Length(Prefix) do
    if Prefix[I] <> Value[I] then Exit;
  Result := True;
end;

function LastBranch(var Revision: ZRevision): PBranch;
begin
  Assert(Length(Revision) mod SizeOf(ZBranch) = 0);
  Result := @Revision[Length(Revision) - SizeOf(ZBranch) + 1];
end;

function BranchRevision(var Revision: ZRevision): Word;
var
  P: PBranch;
begin
  Result := 0;
  if Length(Revision) >= SizeOf(ZBranch) then
  begin
    P := LastBranch(Revision);
    if P <> nil then Result := P.Minor;
  end;
end;

function TZlibArchive.GetZLibBranchTip(Revision: ZRevision): ZRevision;
var
  MustMatch: ZRevision;
  Len: Integer;
  BranchCount: Integer;
  I: Integer;
  DirEntry: TZDirEntry;
begin
  Assert(Length(Revision) mod SizeOf(ZBranch) = 0);
  BranchCount := Length(Revision) div SizeOf(ZBranch);
  if Length(Revision) div SizeOf(ZBranch) < 2 then
    MustMatch := '' else
    MustMatch := Copy(Revision, 1, SizeOf(ZBranch) * (BranchCount - 1) +
      SizeOf(Word));
  Len := Length(Revision);
  if Len = 0 then Len := SizeOf(ZBranch);
  Result := '';
  for I := 0 to FDirectory.Count - 1 do
  begin
    DirEntry := FDirectory[I];
    if Length(DirEntry.Revision) = Len then
      if IsPrefix(MustMatch, DirEntry.Revision) and
        (BranchRevision(Result) < BranchRevision(DirEntry.Revision)) then
        Result := DirEntry.Revision;
  end;
  if Result = '' then Result := FNewRevision;
end;

function TZLibArchive.GetVersion: string;
var
  Rev: ZRevision;
begin
  ValidState;
  if FCurrentRevision <> nil then Rev := FCurrentRevision.Revision
  else Rev := FNewRevision;
  Result := ZRevisionToText(Rev);
end;

procedure TZLibArchive.PositionToVersion(const Version: string);
var
  Index: Integer;
begin
  ValidState;
  Index := IndexOfRevision(TextToZRevision(Version));
  if Index = -1 then FCurrentRevision := nil
  else FCurrentRevision := FDirectory[Index];
end;

procedure TZLibArchive.SetLocked(Value: Boolean);
begin
  ValidState;
  if FReadOnly then
    raise Exception.Create(sReadOnlyArchive);
  if FCurrentRevision <> nil then
    FCurrentRevision.Locked := Value;
end;

function TZLibArchive.PutSource(Source: ISource; const Comment: string;
  ForceBranch: Boolean): string;
var
  DirEntry: TZDirEntry;
  Adaptor: TSourceAdaptor;
  NewRev: ZRevision;
  TempName: string;
  TempFile: TStream;
  Compressor: TCompressionStream;
begin
  if FReadOnly then
    raise Exception.Create(sReadOnlyArchive);

  Result := '';

  // get the next revision
  NewRev := GetNextRevision(ForceBranch);

  // compress the input source into a temp file
  TempName := GetTemporaryFileName('zlb');

  Adaptor := TSourceAdaptor.Create(Source);
  try

    TempFile := TFileStream.Create(TempName, fmCreate);
    try
      Compressor := TCompressionStream.Create(clMax, TempFile);
      try
        Compressor.CopyFrom(Adaptor, 0);
      finally
        Compressor.Free;
      end;

      //  create the directory entry
      DirEntry := TZDirEntry.Create(FileVersion);
      try

        // calc the next revision number
        DirEntry.Revision := NewRev;

        Result := ZRevisionToText(DirEntry.Revision);

        DirEntry.Offset := FHeader.DirectoryOffset;
        DirEntry.OriginalSize := Source.GetLength;
        DirEntry.StoredSize := TempFile.Size;
        DirEntry.FileDate := UTCDateToLocalDate(Source.GetDate);
        DirEntry.CheckinDate := Now;
        DirEntry.Author := FController.GetUser;
        DirEntry.Comment := Comment;

        // copy the temp file to the data file
        TempFile.Position := 0;
        FStream.Position := FHeader.DirectoryOffset;
        FStream.CopyFrom(TempFile, 0);

        // update the file info
        FDirectory.Add(DirEntry);

        FHeader.DirectoryOffset := FHeader.DirectoryOffset + DirEntry.StoredSize;
        FHeader.RevCount := FHeader.RevCount + 1;
        WriteHeader;
        WriteDirectory;

        if FCurrentRevision <> nil then
          FCurrentRevision.Locked := False;
        FCurrentRevision := DirEntry;

      except
        DirEntry.Free;
      end;

    finally
      TempFile.Free;
      SysUtils.DeleteFile(TempName);
    end;

  finally
    Adaptor.Free;
  end;
end;

function TZLibArchive.CheckIn(const Source: ISource; const Comment: string): string;
begin
  if FReadOnly then
    raise Exception.Create(sReadOnlyArchive);
  Result := PutSource(Source, Comment, False);
end;

function TZLibArchive.GetCurrentRevision: ISource;
var
  TempName: string;
begin
  Result := nil;
  ValidState;
  if FCurrentRevision = nil then
    Exit;

  TempName := GetTemporaryFileName('zlb');
  try
    CopyToFile(TempName);
    Result := TTempFileSource.Create(TempName);
  except
    SysUtils.DeleteFile(TempName);
    raise;
  end;
end;

procedure TZLibArchive.CopyToFile(const Target: string);
var
  Expander: TDeCompressionStream;
  TargetFile: TStream;
begin
  ValidState;
  if FCurrentRevision = nil then
    Exit;

  TargetFile := TFileStream.Create(Target, fmCreate);
  try
    FStream.Position := FCurrentRevision.Offset;

    Expander := TDecompressionStream.Create(FStream);
    try
      TargetFile.CopyFrom(Expander, FCurrentRevision.OriginalSize);
    finally
      Expander.Free;
    end;
    FileSetDate(TFileStream(TargetFile).Handle,
      DateTimeToFileDate(FCurrentRevision.FileDate));
  finally
    TargetFile.Free;
  end;
end;

function TZLibArchive.CheckOut: ISource;
begin
  // "locks" the revision
  if FReadOnly then
    raise Exception.Create(sReadOnlyArchive);

  Result := GetCurrentRevision;
  if Result <> nil then FCurrentRevision.Locked := True;
end;

function TZLibArchive.Source: ISource;
begin
  // doesn't "lock" the revision
  Result := GetCurrentRevision;
end;

procedure TZLibArchive.WriteHeader;
begin
  FStream.Position := 0;
  FHeader.Version := FileVersion;
  FStream.Write(FHeader, SizeOf(FHeader));
end;

procedure TZLibArchive.WriteDirectory;
var
  I: Integer;
begin
  FStream.Position := FHeader.DirectoryOffset;
  for I := 0 to FDirectory.Count - 1 do
    TZDirEntry(FDirectory[I]).SaveToStream(FStream);
end;

function TZLibArchive.GetCreator: string;
begin
  Result := FHeader.Creator;
end;

function TZLibArchive.GetCreationDate: TDateTime;
begin
  Result := LocalDateToUTCDate(FHeader.CreationDate);
end;

function TZLibArchive.GetDirectoryCount: Integer;
begin
  Result := FDirectory.Count;
end;

function TZLibArchive.GetDirectoryEntry(I: Integer): TZDirEntry;
begin
  Result := FDirectory[I];
end;

function TZLibArchive.GetNextRevision(ForceBranch: Boolean): ZRevision;
begin
  if FReadOnly then
    raise Exception.Create(sReadOnlyArchive);
  if FCurrentRevision <> nil then
  begin
    if not FCurrentRevision.Locked then
      raise Exception.Create('Archive not Locked');
    Result := FCurrentRevision.Revision
  end else Result := FNewRevision;
  if (Result <> GetZLibBranchTip(Result)) or ForceBranch then
  begin
    Result := Result + FNewRevision;
    while IndexOfRevision(Result) <> - 1 do
      Inc(LastBranch(Result)^.Major);
  end
  else if FCurrentRevision <> nil then Inc(LastBranch(Result)^.Minor);
end;

{ TZLibRevisionHistory }
constructor TZLibRevisionHistory.Create(Controller: TZLibVersionController;
  const ArchiveName: string);
var
  Archive: TZLibArchive;
  I: Integer;
  RevRecord: TZRevRecord;
  Entry: TZDirEntry;
begin
  inherited Create;
  FList := TList.Create;
  FArchiveName := ArchiveName;

  if not FileExists(ArchiveName) then
    Exit;

  Archive := Controller.InternalOpenArchive(ArchiveName);
  try
    FSourceName := Archive.SourceName;
    FCreationDate := Archive.CreationDate;
    FCreator := Archive.Creator;

    for I := Archive.DirectoryCount - 1 downto 0 do
    begin
      Entry := Archive.DirectoryEntry[I];

      RevRecord := TZRevRecord.Create;
      try
        RevRecord.Revision := ZRevisionToText(Entry.Revision);
        RevRecord.ModDate := Entry.FileDate;
        RevRecord.CheckinDate := Entry.CheckinDate;
        RevRecord.Author := Entry.Author;
        RevRecord.Comment := Entry.Comment;
        RevRecord.OriginalSize := Entry.OriginalSize;
        RevRecord.StoredSize := Entry.StoredSize;

        FList.Add(RevRecord);
      except
        RevRecord.Free;
        raise;
      end;
    end;
  finally
    Archive.Free;
  end;
end;

destructor TZLibRevisionHistory.Destroy;
var
  I: Integer;
begin
  if FList <> nil then
  begin
    for I := 0 to FList.Count - 1 do
      TZRevRecord(FList[I]).Free;
    FList.Destroy;
  end;
  inherited Destroy;
end;

function TZLibRevisionHistory.GetRevCount: Integer;
begin
  Result := FList.Count;
end;

function TZLibRevisionHistory.GetRevision(Index: Integer): string;
begin
  Result := TZRevRecord(FList[Index]).Revision;
end;

function TZLibRevisionHistory.GetArchiveName: string;
begin
  Result := FArchiveName;
end;

function TZLibRevisionHistory.GetAuthor(Index: Integer): string;
begin
  Result := TZRevRecord(FList[Index]).Author;
end;

function TZLibRevisionHistory.GetCheckinDate(Index: Integer): TDateTime;
begin
  Result := LocalDateToUTCDate(TZRevRecord(FList[Index]).CheckinDate);
end;

function TZLibRevisionHistory.GetComment(Index: Integer): string;
begin
  Result := TZRevRecord(FList[Index]).Comment;
end;

function TZLibRevisionHistory.GetCreationDate: TDateTime;
begin
  Result := FCreationDate;
end;

function TZLibRevisionHistory.GetCreator: string;
begin
  Result := FCreator;
end;

function TZLibRevisionHistory.GetFileSize(Index: Integer): Integer;
begin
  Result := TZRevRecord(FList[Index]).OriginalSize;
end;

function TZLibRevisionHistory.GetLastModDate(Index: Integer): TDateTime;
begin
  Result := LocalDateToUTCDate(TZRevRecord(FList[Index]).ModDate);
end;

function TZLibRevisionHistory.GetSourceName: string;
begin
  Result := FSourceName;
end;

function TZLibRevisionHistory.GetStoredSize(Index: Integer): Integer;
begin
  Result := TZRevRecord(FList[Index]).StoredSize;
end;


{ ZDirEntry }

constructor TZDirEntry.Create(AFormatVersion: Integer);
begin
  FormatVersion := AFormatVersion;
end;

procedure TZDirEntry.LoadFromStream(S: TStream);
var
  Len: Word;
begin
  S.Read(Len, SizeOf(Len));
  SetLength(Revision, Len * SizeOf(ZBranch));
  S.Read(Revision[1], Len * SizeOf(ZBranch));

  S.Read(Offset, SizeOf(Offset));
  S.Read(OriginalSize, SizeOf(OriginalSize));
  S.Read(StoredSize, SizeOf(StoredSize));
  S.Read(FileDate, SizeOf(FileDate));
  S.Read(CheckInDate, SizeOf(CheckInDate));

  S.Read(Len, SizeOf(Len));
  SetLength(Author, Len);
  S.Read(Author[1], Len);

  S.Read(Len, SizeOf(Len));
  SetLength(Comment, Len);
  S.Read(Comment[1], Len);
end;

procedure TZDirEntry.SaveToStream(S: TStream);
var
  Len: Word;
begin
  Len := Length(Revision) div SizeOf(ZBranch);
  S.Write(Len, SizeOf(Len));
  S.Write(Revision[1], Len * SizeOf(ZBranch));

  S.Write(Offset, SizeOf(Offset));
  S.Write(OriginalSize, SizeOf(OriginalSize));
  S.Write(StoredSize, SizeOf(StoredSize));
  S.Write(FileDate, SizeOf(FileDate));
  S.Write(CheckInDate, SizeOf(CheckInDate));

  Len := Length(Author);
  S.Write(Len, SizeOf(Len));
  S.Write(Author[1], Len);

  Len := Length(Comment);
  S.Write(Len, SizeOf(Len));
  S.Write(Comment[1], Len);
end;



end.
